cdef class Session:
    """
    Session context.
    """

    def __cinit__(self, Session session: Session | None = None):
        cinit(self, session)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Session): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_default(self)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> Session:
        """
        Create object from MPI handle.
        """
        return fromhandle(<MPI_Session> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Call `Finalize` if not null.
        """
        safefree(self)

    #

    @classmethod
    def Init(
        cls,
        Info info: Info = INFO_NULL,
        Errhandler errhandler: Errhandler | None = None,
    ) -> Self:
        """
        Create a new session.
        """
        cdef MPI_Errhandler cerrhdl = arg_Errhandler(errhandler)
        cdef Session session = <Session>New(cls)
        CHKERR( MPI_Session_init(
            info.ob_mpi, cerrhdl, &session.ob_mpi) )
        options_set_errhandler(session.ob_mpi)
        return session

    def Finalize(self) -> None:
        """
        Finalize a session.
        """
        cdef MPI_Session save = self.ob_mpi
        CHKERR( MPI_Session_finalize(&self.ob_mpi) )
        if constobj(self): self.ob_mpi = save

    def Get_num_psets(self, Info info: Info = INFO_NULL) -> int:
        """
        Number of available process sets.
        """
        cdef int num_psets = -1
        CHKERR( MPI_Session_get_num_psets(
            self.ob_mpi, info.ob_mpi, &num_psets) )
        return num_psets

    def Get_nth_pset(self, int n: int, Info info: Info = INFO_NULL) -> str:
        """
        Name of the *n*-th process set.
        """
        cdef int nlen = MPI_MAX_PSET_NAME_LEN
        cdef char *pset_name = NULL
        cdef unused = allocate(nlen+1, sizeof(char), &pset_name)
        CHKERR( MPI_Session_get_nth_pset(
            self.ob_mpi, info.ob_mpi, n, &nlen, pset_name) )
        return mpistr(pset_name)

    def Get_info(self) -> Info:
        """
        Return the current hints for a session.
        """
        cdef Info info = <Info>New(Info)
        CHKERR( MPI_Session_get_info(
            self.ob_mpi, &info.ob_mpi) )
        return info

    def Get_pset_info(self, pset_name: str) -> Info:
        """
        Return the current hints for a session and process set.
        """
        cdef char *cname = NULL
        pset_name = asmpistr(pset_name, &cname)
        cdef Info info = <Info>New(Info)
        CHKERR( MPI_Session_get_pset_info(
            self.ob_mpi, cname, &info.ob_mpi) )
        return info

    def Create_group(self, pset_name: str) -> Group:
        """
        Create a new group from session and process set.
        """
        cdef char *cname = NULL
        pset_name = asmpistr(pset_name, &cname)
        cdef Group group = <Group>New(Group)
        CHKERR( MPI_Group_from_session_pset(
            self.ob_mpi, cname, &group.ob_mpi) )
        return group

    # Buffer Allocation and Usage
    # ---------------------------

    def Attach_buffer(self, buf: Buffer | None) -> None:
        """
        Attach a user-provided buffer for sending in buffered mode.
        """
        cdef void *base = NULL
        cdef MPI_Count size = 0
        buf = attach_buffer(buf, &base, &size)
        with nogil: CHKERR( MPI_Session_attach_buffer_c(
            self.ob_mpi, base, size) )
        detach_buffer_set(self, buf)  # ~> MPI-4.1

    def Detach_buffer(self) -> Buffer | None:
        """
        Remove an existing attached buffer.
        """
        cdef void *base = NULL
        cdef MPI_Count size = 0
        with nogil: CHKERR( MPI_Session_detach_buffer_c(
            self.ob_mpi, &base, &size) )
        return detach_buffer_get(self, base, size)  # ~> MPI-4.1

    def Flush_buffer(self) -> None:
        """
        Block until all buffered messages have been transmitted.
        """
        with nogil: CHKERR( MPI_Session_flush_buffer(self.ob_mpi) )

    def Iflush_buffer(self) -> Request:
        """
        Nonblocking flush for buffered messages.
        """
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_Session_iflush_buffer(
            self.ob_mpi, &request.ob_mpi) )
        return request  # ~> MPI-4.1

    # Error handling
    # --------------

    @classmethod
    def Create_errhandler(
        cls,
        errhandler_fn: Callable[[Session, int], None],
    ) -> Errhandler:
        """
        Create a new error handler for sessions.
        """
        cdef Errhandler errhandler = <Errhandler>New(Errhandler)
        cdef MPI_Session_errhandler_function *fn = NULL
        cdef int index = errhdl_new(errhandler_fn, &fn)
        try:
            CHKERR( MPI_Session_create_errhandler(fn, &errhandler.ob_mpi) )
        except:                     # ~> uncovered  # noqa
            errhdl_del(&index, fn)  # ~> uncovered
            raise                   # ~> uncovered
        return errhandler

    def Get_errhandler(self) -> Errhandler:
        """
        Get the error handler for a session.
        """
        cdef Errhandler errhandler = <Errhandler>New(Errhandler)
        CHKERR( MPI_Session_get_errhandler(self.ob_mpi, &errhandler.ob_mpi) )
        return errhandler

    def Set_errhandler(self, Errhandler errhandler: Errhandler) -> None:
        """
        Set the error handler for a session.
        """
        CHKERR( MPI_Session_set_errhandler(self.ob_mpi, errhandler.ob_mpi) )

    def Call_errhandler(self, int errorcode: int) -> None:
        """
        Call the error handler installed on a session.
        """
        CHKERR( MPI_Session_call_errhandler(self.ob_mpi, errorcode) )

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_Session_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> Session:
        """
        Translate integer handle to object.
        """
        return fromhandle(MPI_Session_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Session_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Session:
        """
        """
        return fromhandle(MPI_Session_f2c(arg))


cdef Session __SESSION_NULL__ = def_Session( MPI_SESSION_NULL , "SESSION_NULL" )


# Predefined session handle
# -------------------------

SESSION_NULL  = __SESSION_NULL__  #: Null session handler



import operator
from functools import reduce

import mpiunittest as unittest

from mpi4py import MPI


def cumsum(seq):
    return reduce(operator.add, seq, 0)


def cumprod(seq):
    return reduce(operator.mul, seq, 1)


_basic = [
    None,
    True,
    False,
    -7,
    0,
    7,
    2**31,
    -(2**63) + 1,
    2**63 - 1,
    -2.17,
    0.0,
    +1.62,
    1 + 2j,
    2 - 3j,
    "mpi4py",
]
messages = list(_basic)
messages += [
    list(_basic),
    tuple(_basic),
    {f"k{k}": v for k, v in enumerate(_basic)},
]


@unittest.skipMPI("openmpi(<1.6.0)")
@unittest.skipMPI("msmpi", MPI.COMM_WORLD.Get_size() >= 3)
@unittest.skipMPI("MPICH1")
@unittest.skipIf(MPI.ROOT == MPI.PROC_NULL, "mpi-root")
@unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, "mpi-world-size<2")
class BaseTestCCOObjInter:
    #
    BASECOMM = MPI.COMM_NULL
    INTRACOMM = MPI.COMM_NULL
    INTERCOMM = MPI.COMM_NULL

    def setUp(self):
        size = self.BASECOMM.Get_size()
        rank = self.BASECOMM.Get_rank()
        if rank < size // 2:
            self.COLOR = 0
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = size // 2
        else:
            self.COLOR = 1
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = 0
        self.INTRACOMM = self.BASECOMM.Split(self.COLOR, key=0)
        Create_intercomm = MPI.Intracomm.Create_intercomm
        self.INTERCOMM = Create_intercomm(
            self.INTRACOMM,
            self.LOCAL_LEADER,
            self.BASECOMM,
            self.REMOTE_LEADER,
        )

    def tearDown(self):
        self.INTRACOMM.Free()
        self.INTERCOMM.Free()

    @unittest.skipMPI("MPICH2(<1.0.8)")
    def testBarrier(self):
        self.INTERCOMM.barrier()

    def testBcast(self):
        rank = self.INTERCOMM.Get_rank()
        size = self.INTERCOMM.Get_size()
        rsize = self.INTERCOMM.Get_remote_size()
        for smess in [*messages, messages]:
            for color in [0, 1]:
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            rmess = self.INTERCOMM.bcast(smess, root=MPI.ROOT)
                        else:
                            rmess = self.INTERCOMM.bcast(
                                None, root=MPI.PROC_NULL
                            )
                        self.assertIsNone(rmess)
                else:
                    for root in range(rsize):
                        rmess = self.INTERCOMM.bcast(None, root=root)
                        self.assertEqual(rmess, smess)

    def testGather(self):
        rank = self.INTERCOMM.Get_rank()
        size = self.INTERCOMM.Get_size()
        rsize = self.INTERCOMM.Get_remote_size()
        for smess in [*messages, messages]:
            for color in [0, 1]:
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            rmess = self.INTERCOMM.gather(smess, root=MPI.ROOT)
                            self.assertEqual(rmess, [smess] * rsize)
                        else:
                            rmess = self.INTERCOMM.gather(
                                None, root=MPI.PROC_NULL
                            )
                            self.assertIsNone(rmess)
                else:
                    for root in range(rsize):
                        rmess = self.INTERCOMM.gather(smess, root=root)
                        self.assertIsNone(rmess)

    @unittest.skipMPI("msmpi(<8.0.0)")
    def testScatter(self):
        rank = self.INTERCOMM.Get_rank()
        size = self.INTERCOMM.Get_size()
        rsize = self.INTERCOMM.Get_remote_size()
        for smess in [*messages, messages]:
            for color in [0, 1]:
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            rmess = self.INTERCOMM.scatter(
                                [smess] * rsize, root=MPI.ROOT
                            )
                        else:
                            rmess = self.INTERCOMM.scatter(
                                None, root=MPI.PROC_NULL
                            )
                        self.assertIsNone(rmess)
                else:
                    for root in range(rsize):
                        rmess = self.INTERCOMM.scatter(None, root=root)
                        self.assertEqual(rmess, smess)

    @unittest.skipMPI("MPICH2(<1.0.8)")
    def testAllgather(self):
        self.INTERCOMM.Get_rank()
        self.INTERCOMM.Get_size()
        rsize = self.INTERCOMM.Get_remote_size()
        for smess in [*messages, messages]:
            rmess = self.INTERCOMM.allgather(smess)
            self.assertEqual(rmess, [smess] * rsize)

    def testAlltoall(self):
        self.INTERCOMM.Get_rank()
        self.INTERCOMM.Get_size()
        rsize = self.INTERCOMM.Get_remote_size()
        for smess in [*messages, messages]:
            rmess = self.INTERCOMM.alltoall([smess] * rsize)
            self.assertEqual(rmess, [smess] * rsize)

    def testReduce(self):
        rank = self.INTERCOMM.Get_rank()
        size = self.INTERCOMM.Get_size()
        rsize = self.INTERCOMM.Get_remote_size()
        for op in (MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD):
            for color in [0, 1]:
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            value = self.INTERCOMM.reduce(
                                None, op=op, root=MPI.ROOT
                            )
                            if op == MPI.SUM:
                                self.assertEqual(value, cumsum(range(rsize)))
                            elif op == MPI.PROD:
                                self.assertEqual(value, cumprod(range(rsize)))
                            elif op == MPI.MAX:
                                self.assertEqual(value, rsize - 1)
                            elif op == MPI.MIN:
                                self.assertEqual(value, 0)
                        else:
                            value = self.INTERCOMM.reduce(
                                None, op=op, root=MPI.PROC_NULL
                            )
                            self.assertIsNone(value)
                else:
                    for root in range(rsize):
                        value = self.INTERCOMM.reduce(rank, op=op, root=root)
                        self.assertIsNone(value)
        badroots = {-3, -2, -1, rsize}.difference({MPI.ROOT, MPI.PROC_NULL})
        for root in badroots:
            with self.assertRaises(MPI.Exception) as cm:
                self.INTERCOMM.reduce(None, op=MPI.NO_OP, root=root)
            ierr = cm.exception.Get_error_class()
            self.assertEqual(ierr, MPI.ERR_ROOT)

    @unittest.skipMPI("MPICH2(<1.0.8)")
    def testAllreduce(self):
        rank = self.INTERCOMM.Get_rank()
        self.INTERCOMM.Get_size()
        rsize = self.INTERCOMM.Get_remote_size()
        for op in (MPI.SUM, MPI.MAX, MPI.MIN, MPI.PROD):
            value = self.INTERCOMM.allreduce(rank, op)
            if op == MPI.SUM:
                self.assertEqual(value, cumsum(range(rsize)))
            elif op == MPI.PROD:
                self.assertEqual(value, cumprod(range(rsize)))
            elif op == MPI.MAX:
                self.assertEqual(value, rsize - 1)
            elif op == MPI.MIN:
                self.assertEqual(value, 0)


class TestCCOObjInter(BaseTestCCOObjInter, unittest.TestCase):
    #
    BASECOMM = MPI.COMM_WORLD


class TestCCOObjInterDup(TestCCOObjInter):
    #
    def setUp(self):
        self.BASECOMM = self.BASECOMM.Dup()
        super().setUp()

    def tearDown(self):
        self.BASECOMM.Free()
        super().tearDown()


class TestCCOObjInterDupDup(TestCCOObjInterDup):
    #
    BASECOMM = MPI.COMM_WORLD
    INTERCOMM_ORIG = MPI.COMM_NULL

    def setUp(self):
        super().setUp()
        self.INTERCOMM_ORIG = self.INTERCOMM
        self.INTERCOMM = self.INTERCOMM.Dup()

    def tearDown(self):
        super().tearDown()
        self.INTERCOMM_ORIG.Free()


if __name__ == "__main__":
    unittest.main()



import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


@unittest.skipIf(MPI.MESSAGE_NULL == MPI.MESSAGE_NO_PROC, "mpi-message")
class TestMessage(unittest.TestCase):
    #
    def testMessageNull(self):
        null = MPI.MESSAGE_NULL
        self.assertFalse(null)
        null2 = MPI.Message()
        self.assertEqual(null, null2)
        null3 = MPI.Message(null)
        self.assertEqual(null, null3)

    def testMessageNoProc(self):
        #
        noproc = MPI.MESSAGE_NO_PROC
        self.assertTrue(noproc)
        noproc.Recv(None)
        self.assertTrue(noproc)
        noproc.Irecv(None).Wait()
        self.assertTrue(noproc)
        #
        noproc2 = MPI.Message(MPI.MESSAGE_NO_PROC)
        self.assertTrue(noproc2)
        self.assertEqual(noproc2, noproc)
        self.assertNotEqual(noproc, MPI.MESSAGE_NULL)
        #
        message = MPI.Message(MPI.MESSAGE_NO_PROC)
        message.Recv(None)
        self.assertEqual(message, MPI.MESSAGE_NULL)
        #
        message = MPI.Message(MPI.MESSAGE_NO_PROC)
        request = message.Irecv(None)
        self.assertEqual(message, MPI.MESSAGE_NULL)
        self.assertNotEqual(request, MPI.REQUEST_NULL)
        request.Wait()
        self.assertEqual(request, MPI.REQUEST_NULL)

    def testPickle(self):
        from pickle import dumps, loads

        for message in (
            MPI.MESSAGE_NULL,
            MPI.MESSAGE_NO_PROC,
        ):
            msg = loads(dumps(message))
            self.assertIs(msg, message)
            msg = loads(dumps(MPI.Message(message)))
            self.assertIsNot(msg, message)
            self.assertEqual(msg, message)
        comm = MPI.COMM_SELF
        request = comm.Isend(b"", 0, 0)
        with self.assertRaises(ValueError):
            loads(dumps(request))
        message = comm.Mprobe(0, 0)
        with self.assertRaises(ValueError):
            loads(dumps(message))
        message.Recv(bytearray(1))
        request.Wait()


@unittest.skipIf(MPI.MESSAGE_NULL == MPI.MESSAGE_NO_PROC, "mpi-message")
class BaseTestP2PMatched:
    #
    COMM = MPI.COMM_NULL

    def testIMProbe(self):
        comm = self.COMM.Dup()
        try:
            m = comm.Improbe()
            self.assertIsNone(m)
            m = comm.Improbe(MPI.ANY_SOURCE)
            self.assertIsNone(m)
            m = comm.Improbe(MPI.ANY_SOURCE, MPI.ANY_TAG)
            self.assertIsNone(m)
            status = MPI.Status()
            m = comm.Improbe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertIsNone(m)
            self.assertEqual(status.source, MPI.ANY_SOURCE)
            self.assertEqual(status.tag, MPI.ANY_TAG)
            self.assertEqual(status.error, MPI.SUCCESS)
            m = MPI.Message.Iprobe(comm)
            self.assertIsNone(m)
            buf = [None, 0, MPI.BYTE]
            s = comm.Isend(buf, comm.rank, 0)
            r = comm.Mprobe(comm.rank, 0).Irecv(buf)
            MPI.Request.Waitall([s, r])
        finally:
            comm.Free()

    def testProbeRecv(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                for s in range(size + 1):
                    with self.subTest(s=s):
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s)
                        if size == 1:
                            n = comm.Improbe(0, 0)
                            self.assertIsNone(n)
                            sr = comm.Isend(sbuf.as_mpi(), 0, 0)
                            m = comm.Mprobe(0, 0)
                            self.assertIsInstance(m, MPI.Message)
                            self.assertTrue(m)
                            rr = m.Irecv(rbuf.as_raw())
                            self.assertFalse(m)
                            self.assertTrue(sr)
                            self.assertTrue(rr)
                            MPI.Request.Waitall([sr, rr])
                            self.assertFalse(sr)
                            self.assertFalse(rr)
                            #
                            n = comm.Improbe(0, 0)
                            self.assertIsNone(n)
                            r = comm.Isend(sbuf.as_mpi(), 0, 0)
                            m = MPI.Message.Probe(comm, 0, 0)
                            self.assertIsInstance(m, MPI.Message)
                            self.assertTrue(m)
                            m.Recv(rbuf.as_raw())
                            self.assertFalse(m)
                            r.Wait()
                            #
                            n = MPI.Message.Iprobe(comm, 0, 0)
                            self.assertIsNone(n)
                            r = comm.Isend(sbuf.as_mpi(), 0, 0)
                            comm.Probe(0, 0)
                            m = MPI.Message.Iprobe(comm, 0, 0)
                            self.assertIsInstance(m, MPI.Message)
                            self.assertTrue(m)
                            m.Recv(rbuf.as_raw())
                            self.assertFalse(m)
                            r.Wait()
                            #
                            n = MPI.Message.Iprobe(comm, 0, 0)
                            self.assertIsNone(n)
                            r = comm.Isend(sbuf.as_mpi(), 0, 0)
                            m = comm.Mprobe(0, 0)
                            self.assertIsInstance(m, MPI.Message)
                            self.assertTrue(m)
                            m.Recv(rbuf.as_raw())
                            self.assertFalse(m)
                            r.Wait()
                        elif rank == 0:
                            n = comm.Improbe(0, 0)
                            self.assertIsNone(n)
                            #
                            comm.Send(sbuf.as_mpi(), 1, 0)
                            m = comm.Mprobe(1, 0)
                            self.assertTrue(m)
                            m.Recv(rbuf.as_raw())
                            self.assertFalse(m)
                            #
                            n = comm.Improbe(0, 0)
                            self.assertIsNone(n)
                            comm.Send(sbuf.as_mpi(), 1, 1)
                            m = None
                            while not m:
                                m = comm.Improbe(1, 1)
                            m.Irecv(rbuf.as_raw()).Wait()
                        elif rank == 1:
                            n = comm.Improbe(1, 0)
                            self.assertIsNone(n)
                            #
                            m = comm.Mprobe(0, 0)
                            self.assertTrue(m)
                            m.Recv(rbuf.as_raw())
                            self.assertFalse(m)
                            #
                            n = comm.Improbe(1, 0)
                            self.assertIsNone(n)
                            comm.Send(sbuf.as_mpi(), 0, 0)
                            m = None
                            while not m:
                                m = comm.Improbe(0, 1)
                            m.Irecv(rbuf.as_mpi()).Wait()
                            comm.Send(sbuf.as_mpi(), 0, 1)
                        else:
                            rbuf = sbuf
                        check = arrayimpl.scalar(s)
                        for value in rbuf:
                            self.assertEqual(value, check)


class TestP2PMatchedSelf(BaseTestP2PMatched, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestP2PMatchedWorld(BaseTestP2PMatched, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestP2PMatchedSelfDup(TestP2PMatchedSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


class TestP2PMatchedWorldDup(TestP2PMatchedWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------

cdef object op_MAX(object x, object y):
    """maximum"""
    if y > x:
        return y
    else:
        return x

cdef object op_MIN(object x, object y):
    """minimum"""
    if y < x:
        return y
    else:
        return x

cdef object op_SUM(object x, object y):
    """sum"""
    return x + y

cdef object op_PROD(object x, object y):
    """product"""
    return x * y

cdef object op_BAND(object x, object y):
    """bit-wise and"""
    return x & y

cdef object op_BOR(object x, object y):
    """bit-wise or"""
    return x | y

cdef object op_BXOR(object x, object y):
    """bit-wise xor"""
    return x ^ y

cdef object op_LAND(object x, object y):
    """logical and"""
    return bool(x) & bool(y)

cdef object op_LOR(object x, object y):
    """logical or"""
    return bool(x) | bool(y)

cdef object op_LXOR(object x, object y):
    """logical xor"""
    return bool(x) ^ bool(y)

cdef object op_MAXLOC(object x, object y):
    """maximum and location"""
    cdef object i, j, u, v
    u, i = x
    v, j = y
    if u > v:
        return u, i
    elif v > u:
        return v, j
    elif j < i:
        return v, j
    else:
        return u, i

cdef object op_MINLOC(object x, object y):
    """minimum and location"""
    cdef object i, j, u, v
    u, i = x
    v, j = y
    if u < v:
        return u, i
    elif v < u:
        return v, j
    elif j < i:
        return v, j
    else:
        return u, i

cdef object op_REPLACE(object x, object y):
    """replace,  (x, y) -> y"""
    <void> x  # unused
    return y

cdef object op_NO_OP(object x, object y):
    """no-op,  (x, y) -> x"""
    <void> y  # unused
    return x

# -----------------------------------------------------------------------------

ctypedef fused op_count_t:
    int
    MPI_Count

ctypedef fused op_usrfn_t:
    MPI_User_function
    MPI_User_function_c

cdef object op_user_lock     = Lock()
cdef list   op_user_registry = [None]*(1+32)


cdef inline object op_user_call_py(int index, object x, object y, object dt):
    return op_user_registry[index](x, y, dt)

cdef inline void op_user_call_mpi(
    int index,
    void *a, void *b, MPI_Count n, MPI_Datatype t,
) noexcept with gil:
    cdef Datatype datatype
    # errors in user-defined reduction operations are unrecoverable
    try:
        datatype = <Datatype>New(Datatype)
        datatype.ob_mpi = t
        try:
            op_user_call_py(index, mpibuf(a, n), mpibuf(b, n), datatype)
        finally:
            datatype.ob_mpi = MPI_DATATYPE_NULL
    except BaseException as exc:                               # ~> uncovered
        PyErr_DisplayException(exc)                            # ~> uncovered
        PySys_WriteStderr(                                     # ~> uncovered
            b"Fatal Python error: %s\n",                       # ~> uncovered
            b"exception in user-defined reduction operation",  # ~> uncovered
        )                                                      # ~> uncovered
        <void>MPI_Abort(MPI_COMM_WORLD, 1)                     # ~> uncovered


cdef inline void op_user_call(
    int index,
    void *a, void *b, MPI_Count count, MPI_Datatype t,
) noexcept nogil:
    # make it abort if Python has finalized
    if not Py_IsInitialized(): <void>MPI_Abort(MPI_COMM_WORLD, 1)
    # make it abort if module cleanup has been done
    if not py_module_alive():  <void>MPI_Abort(MPI_COMM_WORLD, 1)
    # compute the byte-size of memory buffers
    cdef MPI_Count lb=0, extent=0
    <void>MPI_Type_get_extent_c(t, &lb, &extent)
    cdef MPI_Count n = count * extent
    # make the actual GIL-safe Python call
    op_user_call_mpi(index, a, b, n, t)


@cython.callspec("MPIAPI")
cdef void op_user_01(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 1, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_02(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 2, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_03(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 3, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_04(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 4, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_05(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 5, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_06(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 6, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_07(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 7, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_08(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 8, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_09(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call( 9, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_10(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(10, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_11(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(11, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_12(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(12, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_13(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(13, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_14(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(14, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_15(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(15, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_16(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(16, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_17(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(17, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_18(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(18, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_19(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(19, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_20(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(20, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_21(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(21, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_22(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(22, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_23(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(23, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_24(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(24, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_25(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(25, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_26(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(26, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_27(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(27, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_28(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(28, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_29(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(29, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_30(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(30, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_31(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(31, a, b, n[0], t[0])


@cython.callspec("MPIAPI")
cdef void op_user_32(void *a, void *b, op_count_t *n, MPI_Datatype *t) noexcept nogil:
    op_user_call(32, a, b, n[0], t[0])

cdef inline void op_user_map(int index, op_usrfn_t **fn) noexcept nogil:
    if   index ==  1: fn[0] = op_user_01
    elif index ==  2: fn[0] = op_user_02
    elif index ==  3: fn[0] = op_user_03
    elif index ==  4: fn[0] = op_user_04
    elif index ==  5: fn[0] = op_user_05
    elif index ==  6: fn[0] = op_user_06
    elif index ==  7: fn[0] = op_user_07
    elif index ==  8: fn[0] = op_user_08
    elif index ==  9: fn[0] = op_user_09
    elif index == 10: fn[0] = op_user_10
    elif index == 11: fn[0] = op_user_11
    elif index == 12: fn[0] = op_user_12
    elif index == 13: fn[0] = op_user_13
    elif index == 14: fn[0] = op_user_14
    elif index == 15: fn[0] = op_user_15
    elif index == 16: fn[0] = op_user_16
    elif index == 17: fn[0] = op_user_17
    elif index == 18: fn[0] = op_user_18
    elif index == 19: fn[0] = op_user_19
    elif index == 20: fn[0] = op_user_20
    elif index == 21: fn[0] = op_user_21
    elif index == 22: fn[0] = op_user_22
    elif index == 23: fn[0] = op_user_23
    elif index == 24: fn[0] = op_user_24
    elif index == 25: fn[0] = op_user_25
    elif index == 26: fn[0] = op_user_26
    elif index == 27: fn[0] = op_user_27
    elif index == 28: fn[0] = op_user_28
    elif index == 29: fn[0] = op_user_29
    elif index == 30: fn[0] = op_user_30
    elif index == 31: fn[0] = op_user_31
    elif index == 32: fn[0] = op_user_32

cdef inline void op_user_id_set(Op self, int index) noexcept nogil:
    self.flags &= (1U << 24U) - 1U
    self.flags |= <unsigned> index << 24U

cdef inline int op_user_id_get(Op self) noexcept nogil:
    return <int> (self.flags >> 24U)

cdef inline int op_user_id_pop(Op self) noexcept nogil:
    cdef int index = op_user_id_get(self)
    self.flags &= (1U << 24U) - 1U
    return index

cdef inline int op_user_new(
    Op self,
    object function,
    MPI_User_function   **fn_i,
    MPI_User_function_c **fn_c,
) except -1:
    # check whether the function is callable
    function.__call__
    # find a free slot in the registry
    # and register the Python function
    cdef int index = 0
    try:
        with op_user_lock:
            index = op_user_registry.index(None, 1)
            op_user_registry[index] = function
    except ValueError:
        raise RuntimeError(
            "cannot create too many user-defined reduction operations",
        )
    # map slot index to the associated C callback,
    # and return the slot index in the registry
    op_user_map(index, fn_i)
    op_user_map(index, fn_c)
    op_user_id_set(self, index)
    return 0

cdef inline int op_user_cpy(
    Op self,
    Op other,
) except -1:
    cdef int index = op_user_id_get(other)
    op_user_id_set(self, index)
    return 0

cdef inline int op_user_del(
    Op self,
) except -1:
    # free slot in the registry
    cdef int index = op_user_id_pop(self)
    if index > 0:
        with op_user_lock:
            op_user_registry[index] = None
    return 0

# -----------------------------------------------------------------------------

cdef inline object op_call(MPI_Op op, int index, object x, object y):
    if op == MPI_MAX     : return op_MAX(x, y)
    if op == MPI_MIN     : return op_MIN(x, y)
    if op == MPI_SUM     : return op_SUM(x, y)
    if op == MPI_PROD    : return op_PROD(x, y)
    if op == MPI_LAND    : return op_LAND(x, y)
    if op == MPI_BAND    : return op_BAND(x, y)
    if op == MPI_LOR     : return op_LOR(x, y)
    if op == MPI_BOR     : return op_BOR(x, y)
    if op == MPI_LXOR    : return op_LXOR(x, y)
    if op == MPI_BXOR    : return op_BXOR(x, y)
    if op == MPI_MAXLOC  : return op_MAXLOC(x, y)
    if op == MPI_MINLOC  : return op_MINLOC(x, y)
    if op == MPI_REPLACE : return op_REPLACE(x, y)
    if op == MPI_NO_OP   : return op_NO_OP(x, y)
    if index > 0         : return op_user_call_py(index, x, y, None)
    raise ValueError("cannot call user-defined operation")

# -----------------------------------------------------------------------------



import pathlib

import numpy as np

from mpi4py import MPI

x1 = -2.0
x2 = +1.0
y1 = -1.0
y2 = +1.0

w = 600
h = 400
maxit = 255


dirname = pathlib.Path(__file__).resolve().parent
executable = dirname / "mandelbrot-worker.exe"

# spawn worker
worker = MPI.COMM_SELF.Spawn(executable, maxprocs=7)
size = worker.Get_remote_size()

# send parameters
rmsg = np.array([x1, x2, y1, y2], dtype="f")
imsg = np.array([w, h, maxit], dtype="i")
worker.Bcast([rmsg, MPI.REAL], root=MPI.ROOT)
worker.Bcast([imsg, MPI.INTEGER], root=MPI.ROOT)

# gather results
counts = np.empty(size, dtype="i")
indices = np.empty(h, dtype="i")
cdata = np.empty([h, w], dtype="i")
worker.Gather(sendbuf=None, recvbuf=[counts, MPI.INTEGER], root=MPI.ROOT)
worker.Gatherv(
    sendbuf=None, recvbuf=[indices, (counts, None), MPI.INTEGER], root=MPI.ROOT
)
worker.Gatherv(
    sendbuf=None,
    recvbuf=[cdata, (counts * w, None), MPI.INTEGER],
    root=MPI.ROOT,
)

# disconnect worker
worker.Disconnect()

# reconstruct full result
M = np.zeros([h, w], dtype="i")
M[indices, :] = cdata

# eye candy (requires matplotlib)
if 1:
    import contextlib

    with contextlib.suppress(Exception):
        from matplotlib import pyplot as plt

        plt.imshow(M, aspect="equal")
        try:
            plt.nipy_spectral()
        except AttributeError:
            plt.spectral()
        plt.pause(2)



import inspect
import pathlib
import textwrap
from collections import UserList


def is_cyfunction(obj):
    return type(obj).__name__ == "cython_function_or_method"


def is_function(obj):
    return (
        inspect.isbuiltin(obj)
        or is_cyfunction(obj)
        or type(obj) is type(ord)
    )  # fmt: skip


def is_method(obj):
    return (
        inspect.ismethoddescriptor(obj)
        or inspect.ismethod(obj)
        or is_cyfunction(obj)
        or type(obj) in (
            type(str.index),
            type(str.__add__),
            type(str.__new__),
        )
    )  # fmt: skip


def is_classmethod(obj):
    return (
        inspect.isbuiltin(obj)
        or type(obj).__name__ in (
            "classmethod",
            "classmethod_descriptor",
        )
    )  # fmt: skip


def is_staticmethod(obj):
    return type(obj).__name__ == "staticmethod"


def is_datadescr(obj):
    return inspect.isdatadescriptor(obj) and not hasattr(obj, "fget")


def is_property(obj):
    return inspect.isdatadescriptor(obj) and hasattr(obj, "fget")


def is_class(obj):
    return inspect.isclass(obj) or type(obj) is type(int)


class Lines(UserList):
    INDENT = " " * 4
    level = 0

    @property
    def add(self):
        return self

    @add.setter
    def add(self, lines):
        if lines is None:
            return
        if isinstance(lines, str):
            lines = textwrap.dedent(lines).strip().split("\n")
        indent = self.INDENT * self.level
        for line in lines:
            self.append(indent + line)


def signature(obj):
    doc = obj.__doc__
    sig = doc.partition("\n")[0]
    return sig or None


def docstring(obj):
    doc = obj.__doc__
    doc = doc.partition("\n")[2]
    doc = textwrap.dedent(doc).strip()
    doc = f'"""{doc}"""'
    doc = textwrap.indent(doc, Lines.INDENT)
    return doc


def visit_constant(constant):
    name, value = constant
    return f"{name}: Final[{type(value).__name__}] = ..."


def visit_function(function):
    sig = signature(function)
    return f"def {sig}: ..."


def visit_method(method):
    sig = signature(method)
    return f"def {sig}: ..."


def visit_datadescr(datadescr):
    sig = signature(datadescr)
    return f"{sig}"


def visit_property(prop, name=None):
    gname = prop.fget.__name__
    pname = name or gname
    if pname == gname:
        return ["@property", visit_method(prop.fget)]
    else:
        return f"{pname} = property({gname})"


def visit_constructor(cls, name="__init__", args=None):
    init = name == "__init__"
    argname = cls.__name__.lower()
    argtype = cls.__name__
    initarg = args or f"{argname}: {argtype} | None = None"
    selfarg = "self" if init else "cls"
    rettype = "None" if init else "Self"
    arglist = f"{selfarg}, {initarg}"
    sig = f"{name}({arglist}) -> {rettype}"
    return f"def {sig}: ..."


def visit_class(cls, done=None):
    skip = {
        "__doc__",
        "__dict__",
        "__module__",
        "__weakref__",
        "__pyx_vtable__",
        "__str__",
        "__repr__",
        "__lt__",
        "__le__",
        "__ge__",
        "__gt__",
    }
    special = {
        "__len__": ("self", "int", None),
        "__bool__": ("self", "bool", None),
        "__hash__": ("self", "int", None),
        "__int__": ("self", "int", None),
        "__index__": ("self", "int", None),
        "__eq__": ("self", "other: object", "/", "bool", None),
        "__ne__": ("self", "other: object", "/", "bool", None),
        "__buffer__": ("self", "flags: int", "/", "memoryview", (3, 12)),
    }
    constructor = (
        "__new__",
        "__init__",
    )

    override = OVERRIDE.get(cls.__name__, {})
    done = set() if done is None else done
    lines = Lines()

    if cls.__name__ == "Exception":
        skip = skip - {f"__{a}{b}__" for a in "lg" for b in "et"}
        constructor = ()
    if "__hash__" in cls.__dict__:
        if cls.__hash__ is None:
            done.add("__hash__")

    try:

        class sub(cls):
            pass

        final = False
    except TypeError:
        final = True
    if final:
        lines.add = "@final"
    base = cls.__base__
    if base is object:
        lines.add = f"class {cls.__name__}:"
    else:
        lines.add = f"class {cls.__name__}({base.__name__}):"
    lines.level += 1

    for name in constructor:
        if name in done:
            continue
        if name in override:
            done.add(name)
            lines.add = override[name]
            continue
        if name in cls.__dict__:
            done.add(name)
            lines.add = visit_constructor(cls, name)
            continue

    dct = cls.__dict__
    keys = list(dct.keys())
    for name in keys:
        if name in done:
            continue

        if name in override:
            done.add(name)
            lines.add = override[name]
            continue

        if name in skip:
            continue

        if name in special:
            done.add(name)
            *args, retv, py = special[name]
            sig = f"{name}({', '.join(args)}) -> {retv}"
            if py is not None:
                lines.add = f"if sys.version_info >= {py}:"
                lines.level += 1
            lines.add = f"def {sig}: ..."
            if py is not None:
                lines.level -= 1
            continue

        attr = getattr(cls, name)

        if is_method(attr):
            done.add(name)
            if name == attr.__name__:
                obj = dct[name]
                if is_classmethod(obj):
                    lines.add = "@classmethod"
                elif is_staticmethod(obj):
                    lines.add = "@staticmethod"
                lines.add = visit_method(attr)
            elif True:
                lines.add = f"{name} = {attr.__name__}"
            continue

        if is_datadescr(attr):
            done.add(name)
            lines.add = visit_datadescr(attr)
            continue

        if is_property(attr):
            done.add(name)
            lines.add = visit_property(attr, name)
            continue

    leftovers = [
        name for name in keys if name not in done and name not in skip
    ]
    if leftovers:
        raise RuntimeError(f"leftovers: {leftovers}")

    lines.level -= 1
    return lines


def visit_module(module, done=None):
    skip = {
        "__doc__",
        "__name__",
        "__loader__",
        "__spec__",
        "__file__",
        "__package__",
        "__builtins__",
    }

    done = set() if done is None else done
    lines = Lines()

    keys = list(module.__dict__.keys())
    keys.sort(key=lambda name: name.startswith("_"))

    constants = [
        (name, getattr(module, name))
        for name in keys
        if all((
            name not in done and name not in skip,
            isinstance(getattr(module, name), int),
        ))
    ]
    for name, value in constants:
        done.add(name)
        if name in OVERRIDE:
            lines.add = OVERRIDE[name]
        else:
            lines.add = visit_constant((name, value))
    if constants:
        lines.add = ""

    for name in keys:
        if name in done or name in skip:
            continue
        value = getattr(module, name)

        if is_class(value):
            done.add(name)
            lines.add = visit_class(value)
            lines.add = ""
            aliases = [
                (k, getattr(module, k))
                for k in keys
                if all((
                    k not in done and k not in skip,
                    getattr(module, k) is value,
                ))
            ]
            for aliasname, target in aliases:
                done.add(aliasname)
                lines.add = f"{aliasname} = {target.__name__}"
            if aliases:
                lines.add = ""
            instances = [
                (k, getattr(module, k))
                for k in keys
                if all((
                    k not in done and k not in skip,
                    type(getattr(module, k)) is value,
                ))
            ]
            for attrname, attrvalue in instances:
                done.add(attrname)
                lines.add = visit_constant((attrname, attrvalue))
            if instances:
                lines.add = ""
            continue

        if is_function(value):
            done.add(name)
            if name == value.__name__:
                lines.add = visit_function(value)
            else:
                lines.add = f"{name} = {value.__name__}"
            continue

    lines.add = ""
    for name in keys:
        if name in done or name in skip:
            continue
        value = getattr(module, name)
        done.add(name)
        if name in OVERRIDE:
            lines.add = OVERRIDE[name]
        else:
            lines.add = visit_constant((name, value))

    leftovers = [
        name for name in keys if name not in done and name not in skip
    ]
    if leftovers:
        raise RuntimeError(f"leftovers: {leftovers}")
    return lines


IMPORTS = """
import sys
from threading import Lock
from typing import (
    Any,
    AnyStr,
    Final,
    Literal,
    NoReturn,
    final,
    overload,
)

if sys.version_info >= (3, 9):
    from collections.abc import (
        Callable,
        Hashable,
        Iterable,
        Iterator,
        Mapping,
        Sequence,
    )
else:
    from typing import (
        Callable,
        Hashable,
        Iterable,
        Iterator,
        Mapping,
        Sequence,
    )
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from os import PathLike
"""

OVERRIDE = {
    "Info": {
        "__iter__": "def __iter__(self) -> Iterator[str]: ...",
        "__getitem__": "def __getitem__(self, item: str, /) -> str: ...",
        "__setitem__": (
            "def __setitem__(self, item: str, value: str, /) -> None: ..."
        ),
        "__delitem__": "def __delitem__(self, item: str, /) -> None: ...",
        "__contains__": "def __contains__(self, value: str, /) -> bool: ...",
    },
    "Op": {
        "__call__": "def __call__(self, x: Any, y: Any, /) -> Any: ...",
    },
    "buffer": {
        "__new__": """
        @overload
        def __new__(cls) -> Self: ...
        @overload
        def __new__(cls, buf: Buffer, /) -> Self: ...
        """,
        "__getitem__": """
        @overload
        def __getitem__(self, item: int, /) -> int: ...
        @overload
        def __getitem__(self, item: slice, /) -> buffer: ...
        """,
        "__setitem__": """
        @overload
        def __setitem__(self, item: int, value: int, /) -> None: ...
        @overload
        def __setitem__(self, item: slice, value: Buffer, /) -> None: ...
        """,
        "__delitem__": None,
    },
    "Pickle": {
        "__new__": None,
        "__init__": """
        @overload
        def __init__(
            self,
            dumps: Callable[[Any, int], bytes],
            loads: Callable[[Buffer], Any],
            protocol: int | None = None,
            threshold: int | None = None,
        ) -> None: ...
        @overload
        def __init__(
            self,
            dumps: Callable[[Any], bytes] | None = None,
            loads: Callable[[Buffer], Any] | None = None,
        ) -> None: ...
        """,
    },
    "__pyx_capi__": "__pyx_capi__: Final[dict[str, Any]] = ...",
    "_typedict": "_typedict: Final[dict[str, Datatype]] = ...",
    "_typedict_c": "_typedict_c: Final[dict[str, Datatype]] = ...",
    "_typedict_f": "_typedict_f: Final[dict[str, Datatype]] = ...",
    "_keyval_registry": None,
}
OVERRIDE.update({
    subtype: {
        "__new__": "def __new__(cls) -> Self: ...",
    }
    for subtype in (
        "BottomType",
        "InPlaceType",
        "BufferAutomaticType",
    )
})  # fmt: skip
OVERRIDE.update({
    subtype: {
        "__new__": str.format("""
        def __new__(cls, {}: {} | None = None) -> Self: ...
        """, basetype.lower(), basetype)
    }
    for basetype, subtype in (
        ("Comm", "Comm"),
        ("Comm", "Intracomm"),
        ("Comm", "Topocomm"),
        ("Comm", "Cartcomm"),
        ("Comm", "Graphcomm"),
        ("Comm", "Distgraphcomm"),
        ("Comm", "Intercomm"),
        ("Request", "Request"),
        ("Request", "Prequest"),
        ("Request", "Grequest"),
    )
})  # fmt: skip
OVERRIDE.update({  # python/mypy#15717
    "Group": {
        "Intersect": """
        @classmethod  # python/mypy#15717
        def Intersect(cls, group1: Group, group2: Group) -> Self: ...
        """
    }
})  # fmt: skip

TYPING = """
from .typing import (  # noqa: E402,I001
    Buffer,
    Bottom,
    InPlace,
)
from .typing import (  # noqa: E402
    BufSpec,
    BufSpecB,
    BufSpecV,
    BufSpecW,
    TargetSpec,
)
"""


def visit_mpi4py_MPI():
    from mpi4py import MPI as module

    lines = Lines()
    lines.add = "# ruff: noqa: A001"
    lines.add = "# ruff: noqa: A002"
    lines.add = "# ruff: noqa: E501"
    lines.add = "# ruff: noqa: Q000"
    lines.add = "# fmt: off"
    lines.add = IMPORTS
    lines.add = ""
    lines.add = visit_module(module)
    lines.add = ""
    lines.add = TYPING
    return lines


def generate(filename):
    filename = pathlib.Path(filename)
    filename.parent.mkdir(parents=True, exist_ok=True)
    with filename.open("w", encoding="utf-8") as f:
        script = pathlib.Path(__file__).name
        print(f"# Generated with `python conf/{script}`", file=f)
        for line in visit_mpi4py_MPI():
            print(line, file=f)


TOPDIR = pathlib.Path(__file__).resolve().parent.parent
OUTDIR = TOPDIR / "src" / "mpi4py"

if __name__ == "__main__":
    generate(OUTDIR / "MPI.pyi")



cdef class Info:
    """
    Info object.
    """

    def __cinit__(self, Info info: Info | None = None):
        cinit(self, info)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Info): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_Info(self)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> Info:
        """
        Create object from MPI handle.
        """
        return fromhandle(<MPI_Info> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Call `Free` if not null or predefined.
        """
        safefree(self)

    # The Info Object
    # ---------------

    @classmethod
    def Create(
        cls,
        items: (
            Info |
            Mapping[str, str] |
            Iterable[tuple[str, str]] |
            None
        ) = None,
    ) -> Self:
        """
        Create a new info object.
        """
        cdef Info info = <Info>New(cls)
        CHKERR( MPI_Info_create(&info.ob_mpi) )
        if items is None: return info
        cdef object key, value
        try:
            if hasattr(items, 'keys'):
                for key in items.keys():
                    info.Set(key, items[key])
            else:
                for key, value in items:
                    info.Set(key, value)
        except:  # noqa
            CHKERR( MPI_Info_free(&info.ob_mpi) )
            raise
        return info

    @classmethod
    def Create_env(cls, args: Sequence[str] | None = None) -> Self:
        """
        Create a new environment info object.
        """
        cdef int argc = 0
        cdef char **argv = MPI_ARGV_NULL
        cdef Info info = <Info>New(cls)
        args = asarray_argv(args, &argv)
        while argv and argv[argc]: argc += 1
        CHKERR( MPI_Info_create_env(argc, argv, &info.ob_mpi) )
        return info

    def Free(self) -> None:
        """
        Free an info object.
        """
        cdef MPI_Info save = self.ob_mpi
        CHKERR( MPI_Info_free(&self.ob_mpi) )
        if constobj(self): self.ob_mpi = save

    def Dup(self) -> Self:
        """
        Duplicate an existing info object.
        """
        cdef Info info = <Info>New(type(self))
        CHKERR( MPI_Info_dup(self.ob_mpi, &info.ob_mpi) )
        return info

    def Get(self, key: str) -> str | None:
        """
        Retrieve the value associated with a key.
        """
        cdef char *ckey = NULL
        cdef char *cvalue = NULL
        cdef int buflen = MPI_MAX_INFO_VAL
        cdef int flag = 0
        key = asmpistr(key, &ckey)
        cdef unused = allocate(buflen+1, sizeof(char), &cvalue)
        CHKERR( MPI_Info_get_string(self.ob_mpi, ckey, &buflen, cvalue, &flag) )
        if not flag: return None
        return mpistr(cvalue)

    def Set(self, key: str, value: str) -> None:
        """
        Store a value associated with a key.
        """
        cdef char *ckey = NULL
        cdef char *cvalue = NULL
        key = asmpistr(key, &ckey)
        value = asmpistr(value, &cvalue)
        CHKERR( MPI_Info_set(self.ob_mpi, ckey, cvalue) )

    def Delete(self, key: str) -> None:
        """
        Remove a (key, value) pair from info.
        """
        cdef char *ckey = NULL
        key = asmpistr(key, &ckey)
        CHKERR( MPI_Info_delete(self.ob_mpi, ckey) )

    def Get_nkeys(self) -> int:
        """
        Return the number of currently defined keys in info.
        """
        cdef int nkeys = 0
        CHKERR( MPI_Info_get_nkeys(self.ob_mpi, &nkeys) )
        return nkeys

    def Get_nthkey(self, int n: int) -> str:
        """
        Return the *n*-th defined key in info.
        """
        cdef char ckey[MPI_MAX_INFO_KEY+1]
        ckey[0] = 0  # just in case
        CHKERR( MPI_Info_get_nthkey(self.ob_mpi, n, ckey) )
        ckey[MPI_MAX_INFO_KEY] = 0  # just in case
        return mpistr(ckey)

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_Info_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> Info:
        """
        Translate integer handle to object.
        """
        return fromhandle(MPI_Info_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Info_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Info:
        """
        """
        return fromhandle(MPI_Info_f2c(arg))

    # Python mapping emulation
    # ------------------------

    def __len__(self) -> int:
        if not self: return 0
        return self.Get_nkeys()

    def __contains__(self, key: str) -> bool:
        if not self: return False
        cdef char *ckey = NULL
        cdef char cvalue[1]
        cdef int buflen = 0
        cdef int flag = 0
        key = asmpistr(key, &ckey)
        CHKERR( MPI_Info_get_string(self.ob_mpi, ckey, &buflen, cvalue, &flag) )
        return <bint>flag

    def __iter__(self) -> Iterator[str]:
        return iter(self.keys())

    def __getitem__(self, key: str) -> str:
        if not self: raise KeyError(key)
        cdef object value = self.Get(key)
        if value is None: raise KeyError(key)
        return value

    def __setitem__(self, key: str, value: str) -> None:
        if not self: raise KeyError(key)
        self.Set(key, value)

    def __delitem__(self, key: str) -> None:
        if not self: raise KeyError(key)
        if key not in self: raise KeyError(key)
        self.Delete(key)

    def get(self, key: str, default: str | None = None) -> str | None:
        """Retrieve value by key."""
        if not self: return default
        cdef object value = self.Get(key)
        if value is None: return default
        return value

    def keys(self) -> list[str]:
        """Return list of keys."""
        if not self: return []
        cdef list keys = []
        cdef int nkeys = self.Get_nkeys()
        cdef object key
        for k in range(nkeys):
            key = self.Get_nthkey(k)
            keys.append(key)
        return keys

    def values(self) -> list[str]:
        """Return list of values."""
        if not self: return []
        cdef list values = []
        cdef int nkeys = self.Get_nkeys()
        cdef object key, val
        for k in range(nkeys):
            key = self.Get_nthkey(k)
            val = self.Get(key)
            values.append(val)
        return values

    def items(self) -> list[tuple[str, str]]:
        """Return list of items."""
        if not self: return []
        cdef list items = []
        cdef int nkeys = self.Get_nkeys()
        cdef object key, value
        for k in range(nkeys):
            key = self.Get_nthkey(k)
            value = self.Get(key)
            items.append((key, value))
        return items

    def update(
        self,
        items: Info | Mapping[str, str] | Iterable[tuple[str, str]] = (),
        **kwds: str,
    ) -> None:
        """Update contents."""
        if not self: raise KeyError
        cdef object key, value
        if hasattr(items, 'keys'):
            for key in items.keys():
                self.Set(key, items[key])
        else:
            for key, value in items:
                self.Set(key, value)
        for key, value in kwds.items():
            self.Set(key, value)

    def pop(self, key: str, *default: str) -> str:
        """Pop value by key."""
        cdef object value = None
        if self:
            value = self.Get(key)
        if value is not None:
            self.Delete(key)
            return value
        if default:
            value, = default
            return value
        raise KeyError(key)

    def popitem(self) -> tuple[str, str]:
        """Pop first item."""
        if not self: raise KeyError
        cdef object key, value
        cdef int nkeys = self.Get_nkeys()
        if nkeys == 0: raise KeyError
        key = self.Get_nthkey(nkeys - 1)
        value = self.Get(key)
        self.Delete(key)
        return (key, value)

    def copy(self) -> Self:
        """Copy contents."""
        if not self: return <Info>New(type(self))
        return self.Dup()

    def clear(self) -> None:
        """Clear contents."""
        if not self: return None
        cdef object key
        cdef int k = 0, nkeys = self.Get_nkeys()
        while k < nkeys:
            key = self.Get_nthkey(0)
            self.Delete(key)
            k += 1


cdef Info __INFO_NULL__ = def_Info( MPI_INFO_NULL , "INFO_NULL" )
cdef Info __INFO_ENV__  = def_Info( MPI_INFO_ENV  , "INFO_ENV"  )


# Predefined info handles
# -----------------------

INFO_NULL = __INFO_NULL__  #: Null info handle
INFO_ENV  = __INFO_ENV__   #: Environment info handle



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""`multiprocessing.pool` interface via `mpi4py.futures`."""

import functools as _functools
import operator as _operator
import os as _os
import threading as _threading
import warnings as _warnings
import weakref as _weakref

from .. import futures as _futures

__all__ = [
    "Pool",
    "ThreadPool",
    "AsyncResult",
    "ApplyResult",
    "MapResult",
]


class Pool:
    """Pool using MPI processes as workers."""

    Executor = _futures.MPIPoolExecutor

    def __init__(
        self,
        processes=None,
        initializer=None,
        initargs=(),
        **kwargs,
    ):
        """Initialize a new Pool instance.

        Args:
            processes: Number of worker processes.
            initializer: An callable used to initialize workers processes.
            initargs: A tuple of arguments to pass to the initializer.

        .. note::

           Additional keyword arguments are passed down to the
           `MPIPoolExecutor` constructor.

        .. warning::

           The *maxtasksperchild* and *context* arguments of
           `multiprocessing.pool.Pool` are not supported.  Specifying
           *maxtasksperchild* or *context* with a value other than `None` will
           issue a warning of category `UserWarning`.

        """
        if processes is not None:
            if processes < 1:
                raise ValueError("number of processes must be at least 1")
        if initializer is not None:
            if not callable(initializer):
                raise TypeError("initializer must be a callable")
        for name in ("maxtasksperchild", "context"):
            if kwargs.pop(name, None) is not None:
                message = f"argument {name!r} is not supported"
                _warnings.warn(message, stacklevel=2)

        self.executor = self.Executor(
            processes,
            initializer=initializer,
            initargs=initargs,
            **kwargs,
        )

    def apply(
        self,
        func,
        args=(),
        kwds={},  # noqa: B006
    ):
        """Call *func* with arguments *args* and keyword arguments *kwds*.

        Equivalent to ``func(*args, **kwds)``.
        """
        # pylint: disable=dangerous-default-value
        future = self.executor.submit(func, *args, **kwds)
        return future.result()

    def apply_async(
        self,
        func,
        args=(),
        kwds={},  # noqa: B006
        callback=None,
        error_callback=None,
    ):
        """Asynchronous version of `apply()` returning `ApplyResult`."""
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        # pylint: disable=dangerous-default-value
        future = self.executor.submit(func, *args, **kwds)
        return ApplyResult(future, callback, error_callback)

    def map(
        self,
        func,
        iterable,
        chunksize=None,
    ):
        """Apply *func* to each element in *iterable*.

        Equivalent to ``list(map(func, iterable))``.

        Block until all results are ready and return them in a `list`.

        The *iterable* is choped into a number of chunks which are submitted as
        separate tasks. The (approximate) size of these chunks can be specified
        by setting *chunksize* to a positive integer.

        Consider using `imap()` or `imap_unordered()` with explicit *chunksize*
        for better efficiency.

        """
        iterable, chunksize = _chunksize(self, iterable, chunksize)
        return list(self.imap(func, iterable, chunksize=chunksize))

    def map_async(
        self,
        func,
        iterable,
        chunksize=None,
        callback=None,
        error_callback=None,
    ):
        """Asynchronous version of `map()` returning `MapResult`."""
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        iterable, chunksize = _chunksize(self, iterable, chunksize)
        result_iterator = self.imap(func, iterable, chunksize=chunksize)
        future = _async_executor(self).submit(list, result_iterator)
        return MapResult(future, callback, error_callback)

    def imap(
        self,
        func,
        iterable,
        chunksize=1,
    ):
        """Like `map()` but return an `iterator`.

        Equivalent to ``map(func, iterable)``.

        """
        return self.executor.map(func, iterable, chunksize=chunksize)

    def imap_unordered(
        self,
        func,
        iterable,
        chunksize=1,
    ):
        """Like `imap()` but ordering of results is arbitrary."""
        return self.executor.map(
            func, iterable, chunksize=chunksize, unordered=True
        )

    def starmap(
        self,
        func,
        iterable,
        chunksize=None,
    ):
        """Apply *func* to each argument tuple in *iterable*.

        Equivalent to ``list(itertools.starmap(func, iterable))``.

        Block until all results are ready and return them in a `list`.

        The *iterable* is choped into a number of chunks which are submitted as
        separate tasks. The (approximate) size of these chunks can be specified
        by setting *chunksize* to a positive integer.

        Consider using `istarmap()` or `istarmap_unordered()` with explicit
        *chunksize* for better efficiency.

        """
        iterable, chunksize = _chunksize(self, iterable, chunksize)
        return list(self.istarmap(func, iterable, chunksize=chunksize))

    def starmap_async(
        self,
        func,
        iterable,
        chunksize=None,
        callback=None,
        error_callback=None,
    ):
        """Asynchronous version of `starmap()` returning `MapResult`."""
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        iterable, chunksize = _chunksize(self, iterable, chunksize)
        result_iterator = self.istarmap(func, iterable, chunksize=chunksize)
        future = _async_executor(self).submit(list, result_iterator)
        return MapResult(future, callback, error_callback)

    def istarmap(
        self,
        func,
        iterable,
        chunksize=1,
    ):
        """Like `starmap()` but return an `iterator`.

        Equivalent to ``itertools.starmap(func, iterable)``.

        """
        return self.executor.starmap(
            func,
            iterable,
            chunksize=chunksize,
        )

    def istarmap_unordered(
        self,
        func,
        iterable,
        chunksize=1,
    ):
        """Like `istarmap()` but ordering of results is arbitrary."""
        return self.executor.starmap(
            func,
            iterable,
            chunksize=chunksize,
            unordered=True,
        )

    def close(self):
        """Prevent any more tasks from being submitted to the pool."""
        self.executor.shutdown(wait=False)

    def terminate(self):
        """Stop the worker processes without completing pending tasks."""
        self.executor.shutdown(wait=False, cancel_futures=True)

    def join(self):
        """Wait for the worker processes to exit."""
        self.executor.shutdown(wait=True)
        _async_executor_shutdown(self)

    def __enter__(self):
        """Return pool."""
        return self

    def __exit__(self, *args):
        """Close pool."""
        self.close()


class ThreadPool(Pool):
    """Pool using threads as workers."""

    Executor = _futures.ThreadPoolExecutor


class AsyncResult:
    """Asynchronous result."""

    def __init__(self, future, callback=None, error_callback=None):
        """Wrap a future object and register callbacks."""
        self.future = future
        self._event = _threading.Event()
        done_cb = _functools.partial(
            _async_future_callback,
            event=self._event,
            callback=callback,
            error_callback=error_callback,
        )
        self.future.add_done_callback(done_cb)

    def get(self, timeout=None):
        """Return the result when it arrives.

        If *timeout* is not `None` and the result does not arrive within
        *timeout* seconds then raise `TimeoutError`.

        If the remote call raised an exception then that exception will be
        reraised.

        """
        self.wait(timeout)
        if not self.ready():
            raise TimeoutError
        try:
            return self.future.result()
        except _futures.CancelledError:
            raise TimeoutError from None

    def wait(self, timeout=None):
        """Wait until the result is available or *timeout* seconds pass."""
        self._event.wait(timeout)

    def ready(self):
        """Return whether the call has completed."""
        return self._event.is_set()

    def successful(self):
        """Return whether the call completed without raising an exception.

        If the result is not ready then raise `ValueError`.

        """
        if not self.ready():
            raise ValueError(f"{self!r} not ready")
        try:
            return self.future.exception() is None
        except _futures.CancelledError:
            return False


class ApplyResult(AsyncResult):
    """Result type of `apply_async()`."""


class MapResult(AsyncResult):
    """Result type of `map_async()` and `starmap_async()`."""


def _chunksize(pool, iterable, chunksize):
    if chunksize is None:
        chunksize = 1
        num_workers = getattr(pool.executor, "_max_workers", 0)
        if num_workers > 0:  # pragma: no branch
            num_tasks = _operator.length_hint(iterable)
            if num_tasks == 0:
                iterable = list(iterable)
                num_tasks = len(iterable)
            if num_tasks > 0:
                quot, rem = divmod(num_tasks, num_workers * 4)
                chunksize = quot + bool(rem)
    return iterable, chunksize


def _async_future_callback(future, event, callback, error_callback):
    assert future.done()  # noqa: S101
    try:
        exception = future.exception()
    except _futures.CancelledError:
        exception = TimeoutError()
    try:
        if exception is not None:
            if error_callback is not None:
                error_callback(exception)
        else:
            if callback is not None:
                callback(future.result())
    finally:
        event.set()


_cpu_count = getattr(_os, "process_cpu_count", _os.cpu_count)

_async_executor_lock = _threading.Lock()

_async_executor_cache = _weakref.WeakKeyDictionary()
# type: _weakref.WeakKeyDictionary[Pool, _futures.ThreadPoolExecutor]


def _async_get_max_workers(pool):
    max_workers = getattr(pool, "_async_max_workers", 0)
    return max_workers or min(4, _cpu_count() or 1)


def _async_executor(pool):
    with _async_executor_lock:
        executor = _async_executor_cache.get(pool)
        if executor is None:
            max_workers = _async_get_max_workers(pool)
            executor = _futures.ThreadPoolExecutor(max_workers)
            _async_executor_cache[pool] = executor
    return executor


def _async_executor_shutdown(pool, wait=True):
    with _async_executor_lock:
        executor = _async_executor_cache.pop(pool, None)
    if executor is not None:
        executor.shutdown(wait=wait)



import contextlib
import unittest
from collections import namedtuple

from mpitestutil import import_MPI


class TestCase(unittest.TestCase):
    #
    def assertAlmostEqual(self, first, second):
        num = complex(second) - complex(first)
        den = max(abs(complex(second)), abs(complex(first))) or 1.0
        if abs(num / den) > 1e-2:
            raise self.failureException(f"{first!r} != {second!r}")

    @contextlib.contextmanager
    def catchNotImplementedError(self, version=None, subversion=0):
        try:
            yield
        except NotImplementedError:
            if version is None:
                return
            MPI = import_MPI()
            mpi_version = MPI.Get_version()
            self.assertLess(mpi_version, (version, subversion))


_Version = namedtuple("_Version", ["major", "minor", "patch"])  # noqa: PYI024


def _parse_version(version):
    version = (*tuple(map(int, version.split("."))), 0, 0, 0)
    return _Version(*version[:3])


class _VersionPredicate:
    #
    def __init__(self, versionPredicateStr):
        import re

        re_name = re.compile(r"(?i)^([a-z_]\w*(?:\.[a-z_]\w*)*)(.*)$")
        re_pred = re.compile(r"^(<=|>=|<|>|!=|==)(.*)$")

        def split(item):
            m = re_pred.match(item)
            op, version = m.groups()
            version = _parse_version(version)
            return op, version

        vpstr = versionPredicateStr.replace(" ", "")
        m = re_name.match(vpstr)
        name, plist = m.groups()
        if plist:
            assert plist[0] == "(" and plist[-1] == ")"
            plist = plist[1:-1]
        pred = [split(p) for p in plist.split(",") if p]
        self.name = name
        self.pred = pred

    def __str__(self):
        if self.pred:
            items = [f"{op}{'.'.join(map(str, ver))}" for op, ver in self.pred]
            return f"{self.name}({','.join(items)})"
        else:
            return self.name

    def satisfied_by(self, version):
        from operator import eq, ge, gt, le, lt, ne

        opmap = {"<": lt, "<=": le, ">": gt, ">=": ge, "==": eq, "!=": ne}
        version = _parse_version(version)
        for op, ver in self.pred:
            if not opmap[op](version, ver):
                return False
        return True


def mpi_predicate(predicate):
    MPI = import_MPI()

    def key(s):
        s = s.replace(" ", "")
        s = s.replace("/", "")
        s = s.replace("-", "")
        s = s.replace("Intel", "I")
        s = s.replace("Microsoft", "MS")
        return s.lower()

    vp = _VersionPredicate(key(predicate))
    if vp.name == "mpi":
        name, version = "mpi", MPI.Get_version()
        version = (*version, 0)
    else:
        name, version = MPI.get_vendor()
    if vp.name == key(name):
        x, y, z = version
        if vp.satisfied_by(f"{x}.{y}.{z}"):
            return vp
    return None


def is_mpi(predicate):
    return mpi_predicate(predicate)


def is_mpi_gpu(predicate, array):
    if array.backend in ("cupy", "numba", "dlpack-cupy"):
        if mpi_predicate(predicate):
            return True
    return False


SkipTest = unittest.SkipTest

skip = unittest.skip

skipIf = unittest.skipIf

skipUnless = unittest.skipUnless


def skipMPI(predicate, *conditions, reason=""):
    version = mpi_predicate(predicate)
    if version:
        if not conditions or any(conditions):
            return unittest.skip(str(version))
    return unittest.skipIf(False, reason)


def disable(what, reason):
    return unittest.skip(reason)(what)


def main(*args, **kwargs):
    from main import main

    try:
        main(*args, **kwargs)
    except SystemExit:
        pass



import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


@unittest.skipMPI("openmpi(<1.6.0)")
@unittest.skipMPI("MPICH1")
@unittest.skipIf(MPI.ROOT == MPI.PROC_NULL, "mpi-root")
@unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, "mpi-world-size<2")
class BaseTestCCOVecInter:
    #
    BASECOMM = MPI.COMM_NULL
    INTRACOMM = MPI.COMM_NULL
    INTERCOMM = MPI.COMM_NULL

    def setUp(self):
        size = self.BASECOMM.Get_size()
        rank = self.BASECOMM.Get_rank()
        if rank < size // 2:
            self.COLOR = 0
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = size // 2
        else:
            self.COLOR = 1
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = 0
        self.INTRACOMM = self.BASECOMM.Split(self.COLOR, key=0)
        Create_intercomm = MPI.Intracomm.Create_intercomm
        self.INTERCOMM = Create_intercomm(
            self.INTRACOMM,
            self.LOCAL_LEADER,
            self.BASECOMM,
            self.REMOTE_LEADER,
        )

    def tearDown(self):
        self.INTRACOMM.Free()
        self.INTERCOMM.Free()

    def testGatherv(self):
        comm = self.INTERCOMM
        rank = comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            rbuf = array(-1, typecode, (rsize, root + color))
                            comm.Gatherv(None, rbuf.as_mpi(), root=MPI.ROOT)
                            check = arrayimpl.scalar(root)
                            for value in rbuf.flat:
                                self.assertEqual(value, check)
                        else:
                            comm.Gatherv(None, None, root=MPI.PROC_NULL)
                else:
                    for root in range(rsize):
                        sbuf = array(root, typecode, root + color)
                        comm.Gatherv(sbuf.as_mpi(), None, root=root)

    def testScatterv(self):
        comm = self.INTERCOMM
        rank = comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            sbuf = array(root, typecode, (rsize, root + color))
                            comm.Scatterv(sbuf.as_mpi(), None, root=MPI.ROOT)
                        else:
                            comm.Scatterv(None, None, root=MPI.PROC_NULL)
                else:
                    for root in range(rsize):
                        rbuf = array(root, typecode, root + color)
                        comm.Scatterv(None, rbuf.as_mpi(), root=root)
                        check = arrayimpl.scalar(root)
                        for value in rbuf:
                            self.assertEqual(value, check)

    def testAllgatherv(self):
        comm = self.INTERCOMM
        comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for n in range(size):
                        sbuf = array(n, typecode, color)
                        rbuf = array(-1, typecode, (rsize, n + color))
                        comm.Allgatherv(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)
                else:
                    for n in range(rsize):
                        sbuf = array(n, typecode, n + color)
                        rbuf = array(-1, typecode, (rsize, color))
                        comm.Allgatherv(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)

    def testAlltoallv(self):
        comm = self.INTERCOMM
        comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for n in range(size):
                        sbuf = array(n, typecode, (rsize, (n + 1) * color))
                        rbuf = array(-1, typecode, (rsize, n + 3 * color))
                        comm.Alltoallv(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)
                else:
                    for n in range(rsize):
                        sbuf = array(n, typecode, (rsize, n + 3 * color))
                        rbuf = array(-1, typecode, (rsize, (n + 1) * color))
                        comm.Alltoallv(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)

    def testAlltoallw1(self):
        comm = self.INTERCOMM
        comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for n in range(size):
                        sbuf = array(n, typecode, (rsize, (n + 1) * color))
                        rbuf = array(-1, typecode, (rsize, n + 3 * color))
                        comm.Alltoallv(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)
                else:
                    for n in range(rsize):
                        sbuf = array(n, typecode, (rsize, n + 3 * color))
                        rbuf = array(-1, typecode, (rsize, (n + 1) * color))
                        comm.Alltoallv(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)

    def testAlltoallw2(self):
        comm = self.INTERCOMM
        comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for n in range(5):
                check = arrayimpl.scalar(n)
                sbuf = array(n, typecode, (rsize, n))
                rbuf = array(-1, typecode, (rsize, n))
                sdt, rdt = sbuf.mpidtype, rbuf.mpidtype
                sex, rex = sdt.extent, rdt.extent
                sdsp = [i * n * sex for i in range(rsize)]
                rdsp = [i * n * rex for i in range(rsize)]
                smsg = (sbuf.as_raw(), ([n] * rsize, sdsp), [sdt] * rsize)
                rmsg = (rbuf.as_raw(), ([n] * rsize, rdsp), [rdt] * rsize)
                try:
                    comm.Alltoallw(smsg, rmsg)
                except NotImplementedError:
                    self.skipTest("mpi-alltoallw")
                for value in rbuf.flat:
                    self.assertEqual(value, check)


class TestCCOVecInter(BaseTestCCOVecInter, unittest.TestCase):
    #
    BASECOMM = MPI.COMM_WORLD


class TestCCOVecInterDup(TestCCOVecInter):
    #
    def setUp(self):
        self.BASECOMM = self.BASECOMM.Dup()
        super().setUp()

    def tearDown(self):
        self.BASECOMM.Free()
        super().tearDown()


if __name__ == "__main__":
    unittest.main()



import mpiunittest as unittest

from mpi4py import MPI

# -----------------------------------------------------------------------------

_errclasscache = {}


def ErrClassName(ierr):
    if not _errclasscache:
        from mpi4py import MPI

        _errclasscache[MPI.SUCCESS] = "SUCCESS"
        for entry in dir(MPI):
            if entry.startswith("ERR_"):
                errcls = getattr(MPI, entry)
                _errclasscache[errcls] = entry
    return _errclasscache.get(ierr, "<unknown>")


@unittest.skipMPI("MPICH2")
class BaseTestCase(unittest.TestCase):
    #
    def setUp(self):
        self.errhdl_world = MPI.COMM_WORLD.Get_errhandler()
        MPI.COMM_WORLD.Set_errhandler(MPI.ERRORS_RETURN)
        self.errhdl_self = MPI.COMM_SELF.Get_errhandler()
        MPI.COMM_SELF.Set_errhandler(MPI.ERRORS_RETURN)

    def tearDown(self):
        MPI.COMM_WORLD.Set_errhandler(self.errhdl_world)
        self.errhdl_world.Free()
        MPI.COMM_SELF.Set_errhandler(self.errhdl_self)
        self.errhdl_self.Free()

    def assertRaisesMPI(self, IErrClass, callableObj, *args, **kwargs):
        from mpi4py import MPI

        excClass = MPI.Exception
        try:
            callableObj(*args, **kwargs)
        except NotImplementedError:
            if MPI.Get_version() >= (2, 0):
                raise
            raise self.failureException("raised NotImplementedError") from None
        except excClass as excValue:
            error_class = excValue.Get_error_class()
        else:
            error_class = None
        if error_class is not None:
            if isinstance(IErrClass, (list, tuple, set)):
                match = error_class in IErrClass
            else:
                match = error_class == IErrClass
            if not match:
                if isinstance(IErrClass, (list, tuple, set)):
                    IErrClassName = [ErrClassName(e) for e in IErrClass]
                    IErrClassName = type(IErrClass)(IErrClassName)
                else:
                    IErrClassName = ErrClassName(IErrClass)
                raise self.failureException(
                    f"generated error class "
                    f"is '{ErrClassName(error_class)}' ({error_class}), "
                    f"but expected '{IErrClassName}' ({IErrClass})"
                )
        else:
            raise self.failureException(f"{excClass.__name__} not raised")


# -----------------------------------------------------------------------------


class TestExcDatatypeNull(BaseTestCase):
    #
    def testDup(self):
        self.assertRaisesMPI(
            MPI.ERR_TYPE,
            MPI.DATATYPE_NULL.Dup,
        )

    def testCommit(self):
        self.assertRaisesMPI(
            MPI.ERR_TYPE,
            MPI.DATATYPE_NULL.Commit,
        )

    def testFree(self):
        self.assertRaisesMPI(
            MPI.ERR_TYPE,
            MPI.DATATYPE_NULL.Free,
        )


class TestExcDatatype(BaseTestCase):
    #
    DATATYPES = (
        MPI.BYTE,
        MPI.PACKED,
        MPI.CHAR,
        MPI.WCHAR,
        MPI.SIGNED_CHAR,
        MPI.UNSIGNED_CHAR,
        MPI.SHORT,
        MPI.UNSIGNED_SHORT,
        MPI.INT,
        MPI.UNSIGNED,
        MPI.UNSIGNED_INT,
        MPI.LONG,
        MPI.UNSIGNED_LONG,
        MPI.LONG_LONG,
        MPI.UNSIGNED_LONG_LONG,
        MPI.FLOAT,
        MPI.DOUBLE,
        MPI.LONG_DOUBLE,
        MPI.SHORT_INT,
        MPI.TWOINT,
        MPI.INT_INT,
        MPI.LONG_INT,
        MPI.FLOAT_INT,
        MPI.DOUBLE_INT,
        MPI.LONG_DOUBLE_INT,
    )

    ERR_TYPE = MPI.ERR_TYPE

    @unittest.skipMPI("msmpi")
    def testFreePredefined(self):
        for dtype in self.DATATYPES:
            if dtype == MPI.DATATYPE_NULL:
                continue
            self.assertRaisesMPI(
                self.ERR_TYPE,
                dtype.Free,
            )
            self.assertNotEqual(dtype, MPI.DATATYPE_NULL)

    def testKeyvalInvalid(self):
        for dtype in self.DATATYPES:
            if dtype == MPI.DATATYPE_NULL:
                continue
            try:
                self.assertRaisesMPI(
                    [MPI.ERR_KEYVAL, MPI.ERR_OTHER],
                    dtype.Get_attr,
                    MPI.KEYVAL_INVALID,
                )
            except NotImplementedError:
                self.skipTest("mpi-type-get_attr")


name, version = MPI.get_vendor()
if name == "Open MPI":
    if version < (1, 4, 3):
        TestExcDatatype.DATATYPES = TestExcDatatype.DATATYPES[1:]
        TestExcDatatype.ERR_TYPE = MPI.ERR_INTERN

# -----------------------------------------------------------------------------


@unittest.skipMPI("msmpi(<=4.2.0)")
class TestExcStatus(BaseTestCase):
    #
    def testGetCount(self):
        status = MPI.Status()
        self.assertRaisesMPI(
            MPI.ERR_TYPE,
            status.Get_count,
            MPI.DATATYPE_NULL,
        )

    def testGetElements(self):
        status = MPI.Status()
        self.assertRaisesMPI(
            MPI.ERR_TYPE,
            status.Get_elements,
            MPI.DATATYPE_NULL,
        )

    @unittest.skipMPI("MPICH1")
    def testSetElements(self):
        status = MPI.Status()
        self.assertRaisesMPI(
            MPI.ERR_TYPE,
            status.Set_elements,
            MPI.DATATYPE_NULL,
            0,
        )


# -----------------------------------------------------------------------------


class TestExcRequestNull(BaseTestCase):
    #
    def testFree(self):
        self.assertRaisesMPI(
            MPI.ERR_REQUEST,
            MPI.REQUEST_NULL.Free,
        )

    def testCancel(self):
        self.assertRaisesMPI(
            MPI.ERR_REQUEST,
            MPI.REQUEST_NULL.Cancel,
        )


# -----------------------------------------------------------------------------


class TestExcOpNull(BaseTestCase):
    #
    def testFree(self):
        self.assertRaisesMPI(
            [MPI.ERR_OP, MPI.ERR_ARG],
            MPI.OP_NULL.Free,
        )


class TestExcOp(BaseTestCase):
    #
    OPS = (
        MPI.MAX,
        MPI.MIN,
        MPI.SUM,
        MPI.PROD,
        MPI.LAND,
        MPI.BAND,
        MPI.LOR,
        MPI.BOR,
        MPI.LXOR,
        MPI.BXOR,
        MPI.MAXLOC,
        MPI.MINLOC,
        MPI.REPLACE,
        MPI.NO_OP,
    )

    def testFreePredefined(self):
        for op in self.OPS:
            if op == MPI.OP_NULL:
                continue
            self.assertRaisesMPI(
                [MPI.ERR_OP, MPI.ERR_ARG],
                op.Free,
            )
            self.assertNotEqual(op, MPI.OP_NULL)


# -----------------------------------------------------------------------------


class TestExcInfoNull(BaseTestCase):
    #
    def testFree(self):
        self.assertRaisesMPI(
            [MPI.ERR_INFO, MPI.ERR_ARG],
            MPI.INFO_NULL.Free,
        )

    def testTruth(self):
        self.assertFalse(bool(MPI.INFO_NULL))

    @unittest.skipMPI("msmpi(<8.1.0)")
    def testDup(self):
        self.assertRaisesMPI(
            [MPI.ERR_INFO, MPI.ERR_ARG],
            MPI.INFO_NULL.Dup,
        )

    def testGet(self):
        self.assertRaisesMPI(
            [MPI.ERR_INFO, MPI.ERR_ARG],
            MPI.INFO_NULL.Get,
            "key",
        )

    def testSet(self):
        self.assertRaisesMPI(
            [MPI.ERR_INFO, MPI.ERR_ARG],
            MPI.INFO_NULL.Set,
            "key",
            "value",
        )

    def testDelete(self):
        self.assertRaisesMPI(
            [MPI.ERR_INFO, MPI.ERR_ARG],
            MPI.INFO_NULL.Delete,
            "key",
        )

    def testGetNKeys(self):
        self.assertRaisesMPI(
            [MPI.ERR_INFO, MPI.ERR_ARG],
            MPI.INFO_NULL.Get_nkeys,
        )

    def testGetNthKey(self):
        self.assertRaisesMPI(
            [MPI.ERR_INFO, MPI.ERR_ARG],
            MPI.INFO_NULL.Get_nthkey,
            0,
        )


class TestExcInfo(BaseTestCase):
    #
    def setUp(self):
        super().setUp()
        self.INFO = MPI.Info.Create()

    def tearDown(self):
        self.INFO.Free()
        self.INFO = None
        super().tearDown()

    def testDelete(self):
        self.assertRaisesMPI(
            MPI.ERR_INFO_NOKEY,
            self.INFO.Delete,
            "key",
        )

    def testGetNthKey(self):
        self.assertRaisesMPI(
            [MPI.ERR_INFO_KEY, MPI.ERR_ARG],
            self.INFO.Get_nthkey,
            0,
        )


try:
    MPI.Info.Create().Free()
except NotImplementedError:
    unittest.disable(TestExcInfo, "mpi-info")
    unittest.disable(TestExcInfoNull, "mpi-info")

# -----------------------------------------------------------------------------


class TestExcGroupNull(BaseTestCase):
    #
    def testCompare(self):
        self.assertRaisesMPI(
            MPI.ERR_GROUP,
            MPI.Group.Compare,
            MPI.GROUP_NULL,
            MPI.GROUP_NULL,
        )
        self.assertRaisesMPI(
            MPI.ERR_GROUP,
            MPI.Group.Compare,
            MPI.GROUP_NULL,
            MPI.GROUP_EMPTY,
        )
        self.assertRaisesMPI(
            MPI.ERR_GROUP,
            MPI.Group.Compare,
            MPI.GROUP_EMPTY,
            MPI.GROUP_NULL,
        )

    def testAccessors(self):
        for method in ("Get_size", "Get_rank"):
            self.assertRaisesMPI(
                MPI.ERR_GROUP,
                getattr(MPI.GROUP_NULL, method),
            )


class TestExcGroup(BaseTestCase):
    #
    pass


# -----------------------------------------------------------------------------


class TestExcSessionNull(BaseTestCase):
    #
    def testGetNumPsets(self):
        self.assertRaisesMPI(
            MPI.ERR_SESSION,
            MPI.SESSION_NULL.Get_num_psets,
        )

    def testGetNthPset(self):
        self.assertRaisesMPI(
            MPI.ERR_SESSION,
            MPI.SESSION_NULL.Get_nth_pset,
            0,
        )

    def testGetInfo(self):
        self.assertRaisesMPI(
            MPI.ERR_SESSION,
            MPI.SESSION_NULL.Get_info,
        )

    def testGetPsetInfo(self):
        self.assertRaisesMPI(
            MPI.ERR_SESSION,
            MPI.SESSION_NULL.Get_pset_info,
            "mpi://SELF",
        )

    def testCreateGroup(self):
        self.assertRaisesMPI(
            MPI.ERR_SESSION,
            MPI.SESSION_NULL.Create_group,
            "mpi://SELF",
        )

    def testGetErrhandler(self):
        self.assertRaisesMPI(
            MPI.ERR_SESSION,
            MPI.SESSION_NULL.Get_errhandler,
        )

    def testSetErrhandler(self):
        self.assertRaisesMPI(
            MPI.ERR_SESSION,
            MPI.SESSION_NULL.Set_errhandler,
            MPI.ERRORS_RETURN,
        )


class TestExcSession(BaseTestCase):
    #
    def setUp(self):
        super().setUp()
        self.SESSION = MPI.Session.Init()
        self.SESSION.Set_errhandler(MPI.ERRORS_RETURN)

    def tearDown(self):
        self.SESSION.Finalize()
        self.SESSION = None
        super().tearDown()

    def testGetNthPsetNeg(self):
        self.assertRaisesMPI(
            MPI.ERR_ARG,
            self.SESSION.Get_nth_pset,
            -1,
        )

    @unittest.skipMPI("mpich(<4.1.0)")
    def testGetNthPsetPos(self):
        self.assertRaisesMPI(
            MPI.ERR_ARG,
            self.SESSION.Get_nth_pset,
            self.SESSION.Get_num_psets(),
        )

    def testGetPsetInfo(self):
        self.assertRaisesMPI(
            MPI.ERR_ARG,
            self.SESSION.Get_pset_info,
            "@qerty!#$",
        )

    def testCreateGroup(self):
        self.assertRaisesMPI(
            [MPI.ERR_ARG, MPI.ERR_UNSUPPORTED_OPERATION],
            self.SESSION.Create_group,
            "@qerty!#$",
        )


try:
    MPI.Session.Init().Finalize()
except NotImplementedError:
    unittest.disable(TestExcSessionNull, "mpi-session")
    unittest.disable(TestExcSession, "mpi-session")

# -----------------------------------------------------------------------------


class TestExcCommNull(BaseTestCase):
    #
    ERR_COMM = MPI.ERR_COMM

    def testFree(self):
        self.assertRaisesMPI(
            MPI.ERR_COMM,
            MPI.COMM_NULL.Free,
        )

    def testCompare(self):
        self.assertRaisesMPI(
            self.ERR_COMM,
            MPI.Comm.Compare,
            MPI.COMM_NULL,
            MPI.COMM_NULL,
        )
        self.assertRaisesMPI(
            self.ERR_COMM,
            MPI.Comm.Compare,
            MPI.COMM_SELF,
            MPI.COMM_NULL,
        )
        self.assertRaisesMPI(
            self.ERR_COMM,
            MPI.Comm.Compare,
            MPI.COMM_WORLD,
            MPI.COMM_NULL,
        )
        self.assertRaisesMPI(
            self.ERR_COMM,
            MPI.Comm.Compare,
            MPI.COMM_NULL,
            MPI.COMM_SELF,
        )
        self.assertRaisesMPI(
            self.ERR_COMM,
            MPI.Comm.Compare,
            MPI.COMM_NULL,
            MPI.COMM_WORLD,
        )

    def testAccessors(self):
        for method in (
            "Get_size",
            "Get_rank",
            "Is_inter",
            "Is_intra",
            "Get_group",
            "Get_topology",
        ):
            self.assertRaisesMPI(
                MPI.ERR_COMM,
                getattr(MPI.COMM_NULL, method),
            )

    def testDisconnect(self):
        try:
            self.assertRaisesMPI(
                MPI.ERR_COMM,
                MPI.COMM_NULL.Disconnect,
            )
        except NotImplementedError:
            self.skipTest("mpi-comm-disconnect")

    @unittest.skipMPI("openmpi(<1.4.2)")
    def testGetAttr(self):
        self.assertRaisesMPI(
            MPI.ERR_COMM,
            MPI.COMM_NULL.Get_attr,
            MPI.TAG_UB,
        )

    @unittest.skipMPI("openmpi(<1.4.1)")
    def testGetErrhandler(self):
        self.assertRaisesMPI(
            [MPI.ERR_COMM, MPI.ERR_ARG],
            MPI.COMM_NULL.Get_errhandler,
        )

    def testSetErrhandler(self):
        self.assertRaisesMPI(
            MPI.ERR_COMM,
            MPI.COMM_NULL.Set_errhandler,
            MPI.ERRORS_RETURN,
        )

    def testIntraNull(self):
        comm_null = MPI.Intracomm()
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Dup)
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Create, MPI.GROUP_EMPTY)
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Split, color=0, key=0)

    def testInterNull(self):
        comm_null = MPI.Intercomm()
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Get_remote_group)
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Get_remote_size)
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Dup)
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Create, MPI.GROUP_EMPTY)
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Split, color=0, key=0)
        self.assertRaisesMPI(MPI.ERR_COMM, comm_null.Merge, high=True)


class TestExcComm(BaseTestCase):
    #
    @unittest.skipMPI("MPICH1")
    def testFreeSelf(self):
        errhdl = MPI.COMM_SELF.Get_errhandler()
        try:
            MPI.COMM_SELF.Set_errhandler(MPI.ERRORS_RETURN)
            self.assertRaisesMPI(
                [MPI.ERR_COMM, MPI.ERR_ARG],
                MPI.COMM_SELF.Free,
            )
        finally:
            MPI.COMM_SELF.Set_errhandler(errhdl)
            errhdl.Free()

    @unittest.skipMPI("MPICH1")
    def testFreeWorld(self):
        self.assertRaisesMPI(
            [MPI.ERR_COMM, MPI.ERR_ARG],
            MPI.COMM_WORLD.Free,
        )

    def testKeyvalInvalid(self):
        self.assertRaisesMPI(
            [MPI.ERR_KEYVAL, MPI.ERR_OTHER],
            MPI.COMM_WORLD.Get_attr,
            MPI.KEYVAL_INVALID,
        )


# -----------------------------------------------------------------------------


class TestExcWinNull(BaseTestCase):
    #
    ERRORS = [MPI.ERR_WIN, MPI.ERR_ARG]
    if unittest.is_mpi("msmpi"):
        ERRORS += [MPI.ERR_OTHER]

    def testFree(self):
        self.assertRaisesMPI(
            self.ERRORS,
            MPI.WIN_NULL.Free,
        )

    def testGetErrhandler(self):
        self.assertRaisesMPI(
            self.ERRORS,
            MPI.WIN_NULL.Get_errhandler,
        )

    def testSetErrhandler(self):
        self.assertRaisesMPI(
            self.ERRORS,
            MPI.WIN_NULL.Set_errhandler,
            MPI.ERRORS_RETURN,
        )

    def testCallErrhandler(self):
        self.assertRaisesMPI(
            [MPI.ERR_WIN, MPI.ERR_ARG],
            MPI.WIN_NULL.Call_errhandler,
            0,
        )


class TestExcWin(BaseTestCase):
    #
    def setUp(self):
        super().setUp()
        self.WIN = MPI.Win.Create(None, 1, MPI.INFO_NULL, MPI.COMM_SELF)
        self.WIN.Set_errhandler(MPI.ERRORS_RETURN)

    def tearDown(self):
        self.WIN.Free()
        self.WIN = None
        super().tearDown()

    def testKeyvalInvalid(self):
        self.assertRaisesMPI(
            [MPI.ERR_KEYVAL, MPI.ERR_OTHER],
            self.WIN.Get_attr,
            MPI.KEYVAL_INVALID,
        )


try:
    MPI.Win.Create(None, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(TestExcWin, "mpi-win")
    unittest.disable(TestExcWinNull, "mpi-win")

# -----------------------------------------------------------------------------


class TestExcErrhandlerNull(BaseTestCase):
    #
    def testFree(self):
        self.assertRaisesMPI(
            [MPI.ERR_ERRHANDLER, MPI.ERR_ARG],
            MPI.ERRHANDLER_NULL.Free,
        )

    def testCommSelfSetErrhandler(self):
        self.assertRaisesMPI(
            [MPI.ERR_ERRHANDLER, MPI.ERR_ARG],
            MPI.COMM_SELF.Set_errhandler,
            MPI.ERRHANDLER_NULL,
        )

    def testCommWorldSetErrhandler(self):
        self.assertRaisesMPI(
            [MPI.ERR_ERRHANDLER, MPI.ERR_ARG],
            MPI.COMM_WORLD.Set_errhandler,
            MPI.ERRHANDLER_NULL,
        )


class TestExcErrhandler(BaseTestCase):
    #
    pass


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()

# -----------------------------------------------------------------------------



import sys
import time

from mpi4py.futures import MPICommExecutor

x0 = -2.0
x1 = +2.0
y0 = -1.5
y1 = +1.5

w = 1600
h = 1200

dx = (x1 - x0) / w
dy = (y1 - y0) / h


def julia(x, y):
    c = complex(0, 0.65)
    z = complex(x, y)
    n = 255
    while abs(z) < 3 and n > 1:
        z = z**2 + c
        n -= 1
    return n


def julia_line(k):
    line = bytearray(w)
    y = y1 - k * dy
    for j in range(w):
        x = x0 + j * dx
        line[j] = julia(x, y)
    return line


def plot(image):
    import contextlib
    import warnings

    warnings.simplefilter("ignore", UserWarning)
    try:
        from matplotlib import pyplot as plt
    except ImportError:
        return
    plt.figure()
    plt.imshow(image, aspect="equal", cmap="cubehelix")
    plt.axis("off")
    with contextlib.suppress(Exception):
        plt.draw()
        plt.pause(2)


def test_julia():
    with MPICommExecutor() as executor:
        if executor is None:
            return  # worker process
        tic = time.time()
        image = list(executor.map(julia_line, range(h), chunksize=10))
        toc = time.time()

    print(f"Julia Set {w}x{h} in {toc - tic:.2f} seconds.")
    if len(sys.argv) > 1 and sys.argv[1] == "-plot":
        plot(image)


if __name__ == "__main__":
    test_julia()



import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


class BaseTestP2PBufPart:
    #
    COMM = MPI.COMM_NULL

    def testSelf(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                for s in range(size):
                    for p in range(1, 4):
                        with self.subTest(p=p, s=s):
                            sbuf = array(s, typecode, s * p)
                            rbuf = array(-1, typecode, s * p)
                            sreq = self.COMM.Psend_init(
                                sbuf.as_mpi(), p, rank, 0
                            )
                            rreq = self.COMM.Precv_init(
                                rbuf.as_mpi(), p, rank, 0
                            )
                            for _ in range(3):
                                rreq.Start()
                                for i in range(p):
                                    flag = rreq.Parrived(i)
                                    self.assertFalse(flag)
                                sreq.Start()
                                for i in range(p):
                                    sreq.Pready(i)
                                    for j in range(i + 1, p):
                                        flag = rreq.Parrived(j)
                                        self.assertFalse(flag)
                                for i in range(p):
                                    while not rreq.Parrived(i):
                                        pass
                                    flag = rreq.Parrived(i)
                                    self.assertTrue(flag)
                                rreq.Wait()
                                sreq.Wait()
                                self.assertNotEqual(sreq, MPI.REQUEST_NULL)
                                self.assertNotEqual(rreq, MPI.REQUEST_NULL)
                                check = arrayimpl.scalar(s)
                                for value in rbuf:
                                    self.assertEqual(value, check)
                            rreq.Free()
                            sreq.Free()
                            self.assertEqual(sreq, MPI.REQUEST_NULL)
                            self.assertEqual(rreq, MPI.REQUEST_NULL)

    def testRing(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                for s in range(size):
                    for p in range(1, 4):
                        with self.subTest(p=p, s=s):
                            sbuf = array(s, typecode, s * p)
                            rbuf = array(-1, typecode, s * p)
                            sreq = self.COMM.Psend_init(
                                sbuf.as_mpi(), p, dest, 0
                            )
                            rreq = self.COMM.Precv_init(
                                rbuf.as_mpi(), p, source, 0
                            )
                            for _ in range(3):
                                self.COMM.Barrier()
                                rreq.Start()
                                for i in range(p):
                                    flag = rreq.Parrived(i)
                                    self.assertFalse(flag)
                                self.COMM.Barrier()
                                sreq.Start()
                                for i in range(p):
                                    sreq.Pready(i)
                                self.COMM.Barrier()
                                for i in range(p):
                                    while not rreq.Parrived(i):
                                        pass
                                    flag = rreq.Parrived(i)
                                    self.assertTrue(flag)
                                rreq.Wait()
                                sreq.Wait()
                                self.assertNotEqual(sreq, MPI.REQUEST_NULL)
                                self.assertNotEqual(rreq, MPI.REQUEST_NULL)
                                self.COMM.Barrier()
                                check = arrayimpl.scalar(s)
                                for value in rbuf:
                                    self.assertEqual(value, check)
                            rreq.Free()
                            sreq.Free()
                            self.assertEqual(sreq, MPI.REQUEST_NULL)
                            self.assertEqual(rreq, MPI.REQUEST_NULL)

    def testRingRangeList(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                for s in range(size):
                    for p in range(1, 4):
                        with self.subTest(p=p, s=s):
                            sbuf = array(s, typecode, s * p)
                            rbuf = array(-1, typecode, s * p)
                            sreq = self.COMM.Psend_init(
                                sbuf.as_mpi(), p, dest, 0
                            )
                            rreq = self.COMM.Precv_init(
                                rbuf.as_mpi(), p, source, 0
                            )
                            for case in range(4):
                                self.COMM.Barrier()
                                rreq.Start()
                                for i in range(p):
                                    flag = rreq.Parrived(i)
                                    self.assertFalse(flag)
                                self.COMM.Barrier()
                                sreq.Start()
                                if p > 1 and case % 2 == 0:
                                    sreq.Pready_range(0, p // 2 - 1)
                                    sreq.Pready_range(p // 2, p - 1)
                                else:
                                    sreq.Pready_list(list(range(p // 2)))
                                    sreq.Pready_list(list(range(p // 2, p)))
                                self.COMM.Barrier()
                                for i in range(p):
                                    while not rreq.Parrived(i):
                                        pass
                                    flag = rreq.Parrived(i)
                                    self.assertTrue(flag)
                                rreq.Wait()
                                sreq.Wait()
                                self.assertNotEqual(sreq, MPI.REQUEST_NULL)
                                self.assertNotEqual(rreq, MPI.REQUEST_NULL)
                                self.COMM.Barrier()
                                check = arrayimpl.scalar(s)
                                for value in rbuf:
                                    self.assertEqual(value, check)
                            rreq.Free()
                            sreq.Free()
                            self.assertEqual(sreq, MPI.REQUEST_NULL)
                            self.assertEqual(rreq, MPI.REQUEST_NULL)


class TestP2PBufPartSelf(BaseTestP2PBufPart, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestP2PBufPartWorld(BaseTestP2PBufPart, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestP2PBufPartSelfDup(TestP2PBufPartSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


class TestP2PBufPartWorldDup(TestP2PBufPartWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


def have_feature():
    info = MPI.Get_library_version()
    if "MPICH" in info and "ch3:" in info:
        raise NotImplementedError
    sreq = MPI.COMM_SELF.Psend_init(bytearray(1), 1, 0, 0)
    rreq = MPI.COMM_SELF.Precv_init(bytearray(1), 1, 0, 0)
    sreq.Start()
    rreq.Start()
    sreq.Pready(0)
    rreq.Parrived(0)
    rreq.Wait()
    rreq.Free()
    del rreq
    sreq.Wait()
    sreq.Free()
    del sreq


try:
    have_feature()
except NotImplementedError:
    unittest.disable(BaseTestP2PBufPart, "mpi-p2p-part")

if __name__ == "__main__":
    unittest.main()
